#include "stdafx.h"
#include "tracer.h"


class CMyApp : public CWinApp
{
public:
//Constructors / Destructors
  CMyApp() noexcept;

//Member variables
  bool m_bResolveAddressesToHostnames;

//Methods
  static CString AddressToString(const SOCKADDR* pSockAddr, int nSockAddrLen, int nFlags, UINT* pnSocketPort);
  static CString GetErrorMessage(DWORD dwError);
  static CString RTTAsString(DWORD dwRTT);
  static CString GetIpErrorString(IP_STATUS dwError);

protected:
//Methods
  BOOL InitInstance() override;
};

class CMyCommandLineInfo : public CCommandLineInfo
{
public:
  CMyCommandLineInfo() noexcept;
  void ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast) override;

//Our specific command line options
  CString m_sHost;
  bool m_bResolveAddressesToHostnames;
  CString m_sLocalBoundAddress;
  UCHAR m_nHopCount;
  UCHAR m_nPings;
  DWORD m_dwTimeout;
  bool m_bIPv6;
};

//Class derived to implement Trace Route
class CMyTraceRoute : public CTraceRoute
{
  bool OnSingleHostResult(int nHostNum, const CHostTraceMultiReplyv4& htmr) override;
  bool OnSingleHostResult(int nHostNum, const CHostTraceMultiReplyv6& htmr) override;
};


#pragma warning(suppress: 26426)
CMyApp theApp;


CMyCommandLineInfo::CMyCommandLineInfo() noexcept : m_bResolveAddressesToHostnames{true},
                                                    m_nHopCount{30},
                                                    m_nPings{3},
                                                    m_dwTimeout{5000},
                                                    m_bIPv6{false}
{
}

void CMyCommandLineInfo::ParseParam(LPCTSTR lpszParam, BOOL bFlag, BOOL bLast)
{
  if (bFlag)
  {
    //Form a CString version of the parameter (just makes it easier to parse with)
    CString sParam{lpszParam};

    if (_tcsicmp(lpszParam, _T("d")) == 0)
      m_bResolveAddressesToHostnames = false;
    else if ((sParam.Find(_T("h:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_nHopCount = static_cast<UCHAR>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("p:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_nPings = static_cast<UCHAR>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("w:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_dwTimeout = static_cast<DWORD>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if (_tcsicmp(lpszParam, _T("4")) == 0)
      m_bIPv6 = false;
    else if (_tcsicmp(lpszParam, _T("6")) == 0)
      m_bIPv6 = true;
    else if ((sParam.Find(_T("S:")) == 0) && (sParam.GetLength() > 2))
      m_sLocalBoundAddress = sParam.Right(sParam.GetLength() - 2);
    else
      __super::ParseParam(lpszParam, bFlag, bLast);
  }
  else
    m_sHost = lpszParam;
}


bool CMyTraceRoute::OnSingleHostResult(int nHostNum, const CHostTraceMultiReplyv4& htmr)
{
  if (htmr.dwError == 0)
  {
    CString sHost;
    if (theApp.m_bResolveAddressesToHostnames)
#pragma warning(suppress: 26490)
      sHost = CMyApp::AddressToString(reinterpret_cast<const SOCKADDR*>(&htmr.Address), sizeof(htmr.Address), NI_NAMEREQD, nullptr);
#pragma warning(suppress: 26490)
    CString sIPAddress{CMyApp::AddressToString(reinterpret_cast<const SOCKADDR*>(&htmr.Address), sizeof(htmr.Address), NI_NUMERICHOST, nullptr)};

    if (sHost.GetLength())
    {
      _tprintf(_T("  %d\t%s\t%s\t%s\t%s [%s]\n"), nHostNum, CMyApp::RTTAsString(htmr.minRTT).GetString(), CMyApp::RTTAsString(htmr.avgRTT).GetString(),
               CMyApp::RTTAsString(htmr.maxRTT).GetString(), sHost.GetString(), sIPAddress.GetString());
    }
    else
    {
      _tprintf(_T("  %d\t%s\t%s\t%s\t%s\n"), nHostNum, CMyApp::RTTAsString(htmr.minRTT).GetString(), CMyApp::RTTAsString(htmr.avgRTT).GetString(),
               CMyApp::RTTAsString(htmr.maxRTT).GetString(), sIPAddress.GetString());
    }
  }
  else
  {
    if (htmr.dwError == ERROR_TIMEOUT)
      _tprintf(_T("  %d\t*\t*\t*\tRequest timed out.\n"), nHostNum);
    else
      _tprintf(_T("  %d\t*\t*\t*\tError:%s\n"), nHostNum, CMyApp::GetErrorMessage(htmr.dwError).GetString());
  }
  return true;
}

bool CMyTraceRoute::OnSingleHostResult(int nHostNum, const CHostTraceMultiReplyv6& htmr)
{
  if (htmr.dwError == 0)
  {
    CString sHost;
    if (theApp.m_bResolveAddressesToHostnames)
#pragma warning(suppress: 26490)
      sHost = CMyApp::AddressToString(reinterpret_cast<const SOCKADDR*>(&htmr.Address), sizeof(htmr.Address), NI_NAMEREQD, nullptr);
#pragma warning(suppress: 26490)
    CString sIPAddress{CMyApp::AddressToString(reinterpret_cast<const SOCKADDR*>(&htmr.Address), sizeof(htmr.Address), NI_NUMERICHOST, nullptr)};

    if (sHost.GetLength())
    {
      _tprintf(_T("  %d\t%s\t%s\t%s\t%s [%s]\n"), nHostNum, CMyApp::RTTAsString(htmr.minRTT).GetString(), CMyApp::RTTAsString(htmr.avgRTT).GetString(),
               CMyApp::RTTAsString(htmr.maxRTT).GetString(), sHost.GetString(), sIPAddress.GetString());
    }
    else
    {
      _tprintf(_T("  %d\t%s\t%s\t%s\t%s\n"), nHostNum, CMyApp::RTTAsString(htmr.minRTT).GetString(), CMyApp::RTTAsString(htmr.avgRTT).GetString(),
               CMyApp::RTTAsString(htmr.maxRTT).GetString(), sIPAddress.GetString());
    }
  }
  else
  {
    if (htmr.dwError == ERROR_TIMEOUT)
      _tprintf(_T("  %d\t*\t*\t*\tRequest timed out.\n"), nHostNum);
    else
      _tprintf(_T("  %d\t*\t*\t*\tError:%s\n"), nHostNum, CMyApp::GetErrorMessage(htmr.dwError).GetString());
  }
  return true;
}


CMyApp::CMyApp() noexcept : m_bResolveAddressesToHostnames{false}
{
}

CString CMyApp::AddressToString(const SOCKADDR* pSockAddr, int nSockAddrLen, int nFlags, UINT* pnSocketPort)
{
  //What will be the return value from this function
  CString sSocketAddress;

  int nResult{0};
#ifdef _UNICODE
  CStringW sName;
  nResult = GetNameInfoW(pSockAddr, nSockAddrLen, sName.GetBuffer(NI_MAXHOST), NI_MAXHOST, nullptr, 0, nFlags);
  sName.ReleaseBuffer();
#else
  CStringA sName;
  nResult = getnameinfo(pSockAddr, nSockAddrLen, sName.GetBuffer(NI_MAXHOST), NI_MAXHOST, nullptr, 0, nFlags);
  sName.ReleaseBuffer();
#endif
  if (nResult == 0)
  {
    sSocketAddress = sName;
    if (pnSocketPort != nullptr)
      *pnSocketPort = ntohs(SS_PORT(&pSockAddr));
  }

  return sSocketAddress;
}

CString CMyApp::GetErrorMessage(DWORD dwError)
{
  CString sError;

  //Lookup the error using FormatMessage
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                     nullptr, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};

  if (dwReturn)
  {
    sError = lpBuffer;
    LocalFree(lpBuffer);
  }

  return sError;
}

CString CMyApp::GetIpErrorString(IP_STATUS dwError)
{
  DWORD dwBuffer{0};
  CStringW sUnicodeError;
  ::GetIpErrorString(dwError, nullptr, &dwBuffer);
  ::GetIpErrorString(dwError, sUnicodeError.GetBuffer(dwBuffer), &dwBuffer);
  sUnicodeError.ReleaseBuffer();
  return CString{sUnicodeError};
}

CString CMyApp::RTTAsString(DWORD dwRTT)
{
  CString sMsg;
  if (dwRTT == 0)
    sMsg = _T("<1 ms");
  else
    sMsg.Format(_T("%u ms"), dwRTT);
  return sMsg;
}

BOOL CMyApp::InitInstance()
{
  //Parse the command line
  CMyCommandLineInfo rCmdInfo;
  for (int i{1}; i<__argc; i++)
  {
  #ifdef _UNICODE
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__wargv[i]};
  #else
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__argv[i]};
  #endif
    BOOL bFlag{FALSE};
    const BOOL bLast{(i + 1) == __argc};
#pragma warning(suppress: 26481)
    if (pszParam[0] == _T('-') || pszParam[0] == _T('/'))
    {
      //remove flag specifier
      bFlag = TRUE;
#pragma warning(suppress: 26481)
      ++pszParam;
    }
#pragma warning(suppress: 26486)
    rCmdInfo.ParseParam(pszParam, bFlag, bLast);
  }
  m_bResolveAddressesToHostnames = rCmdInfo.m_bResolveAddressesToHostnames;

  if (rCmdInfo.m_sHost.IsEmpty() || (rCmdInfo.m_nPings == 0))
  {
    _tprintf(_T("TRACER target_name [-d] [-h:Hopcount] [-w:timeout] [-p:pingcount] [-S:BoundAddress] [-4 | -6]\n\n"));
    _tprintf(_T("Options:\n"));
    _tprintf(_T("  -d               Do not resolve addresses to hostnames.\n"));
    _tprintf(_T("  -h:maximum_hops  Maximum number of hops to search for target.\n"));
    _tprintf(_T("  -w:timeout       Wait timeout milliseconds for each reply.\n"));
    _tprintf(_T("  -p:pingcount     The number of pings per host.\n"));
    _tprintf(_T("  -4               Force using IPv6.\n"));
    _tprintf(_T("  -6               Force using IPv4.\n"));
    _tprintf(_T("  -S:BoundAddress  The local address to bind to.\n"));
    return FALSE;
  }

  //Initialize sockets
  WSADATA wsa{};
  if (WSAStartup(MAKEWORD(2, 0), &wsa) != 0)
  {
    _tprintf(_T("Failed to initialize Winsock Stack\n"));
    return FALSE;
  }

  //Print the intro comment
#pragma warning(suppress: 26472)
  _tprintf(_T("\nTracing route to %s\nover a maximum of %d hops:\n\n"), rCmdInfo.m_sHost.GetString(), static_cast<int>(rCmdInfo.m_nHopCount));

  //Do the actual trace route
  CTraceRoute::CReplyv4 trrv4;
  CTraceRoute::CReplyv6 trrv6;
  CMyTraceRoute tr;
  if (rCmdInfo.m_bIPv6)
  {
    if (tr.Tracev6(rCmdInfo.m_sHost, trrv6, rCmdInfo.m_nHopCount, rCmdInfo.m_dwTimeout, rCmdInfo.m_nPings, 32, 0, false, false, rCmdInfo.m_sLocalBoundAddress.GetLength() ? rCmdInfo.m_sLocalBoundAddress : nullptr))
      _tprintf(_T("\nTrace complete.\n"));
    else
      _tprintf(_T("%s\n"), GetErrorMessage(GetLastError()).GetString());
  }
  else
  {
    if (tr.Tracev4(rCmdInfo.m_sHost, trrv4, rCmdInfo.m_nHopCount, rCmdInfo.m_dwTimeout, rCmdInfo.m_nPings, 32, 0, false, false, rCmdInfo.m_sLocalBoundAddress.GetLength() ? rCmdInfo.m_sLocalBoundAddress : nullptr))
      _tprintf(_T("\nTrace complete.\n"));
    else
      _tprintf(_T("%s\n"), GetErrorMessage(GetLastError()).GetString());
  }

  WSACleanup();

  return FALSE;
}
